# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Type, Tuple, List

from thespian.actors import ActorAddress, ActorExitRequest, ChildActorExited, PoisonMessage

from powerapi.actor import Actor, InitializationException
from powerapi.formula import FormulaActor, FormulaValues
from powerapi.dispatch_rule import DispatchRule
from powerapi.utils import Tree
from powerapi.report import Report
from powerapi.message import StartMessage, DispatcherStartMessage, FormulaStartMessage, EndMessage, ErrorMessage, OKMessage
from powerapi.dispatcher.blocking_detector import BlockingDetector
from powerapi.dispatcher.route_table import RouteTable


def _clean_list(id_list):
    """
    return a list where all elements are unique
    """
    id_list.sort()
    r_list = []
    last_element = None
    for x in id_list:
        if x != last_element:
            r_list.append(x)
            last_element = x
    return r_list


def _extract_formula_id(report: Report, dispatch_rule: DispatchRule, primary_dispatch_rule: DispatchRule) -> List[Tuple]:
    """
    Use the dispatcha rule to extract formula_id from the given report.
    Formula id are then mapped to an identifier that match the primary
    report identifier fields

    ex: primary dispatch_rule (sensor, socket, core)
        second  dispatch_rule (sensor)
    The second dispatch_rule need to match with the primary if sensor are
    equal.

    :param powerapi.Report report:                 Report to split
    :param powerapi.DispatchRule dispatch_rule: DispatchRule rule

    :return: List of formula_id associated to a sub-report of report
    :rtype: [tuple]
    """

    # List of tuple (id_report, report)
    id_list = dispatch_rule.get_formula_id(report)
    if dispatch_rule.is_primary:
        return id_list

    def f(formula_id):
        return _match_report_id(formula_id, dispatch_rule, primary_dispatch_rule)

    return _clean_list(list(map(f, id_list)))


def _match_report_id(report_id: Tuple, dispatch_rule: DispatchRule, primary_rule: DispatchRule) -> Tuple:
    """
    Return the new_report_id with the report_id by removing
    every "useless" fields from it.

    :param tuple report_id:                     Original report id
    :param powerapi.DispatchRule dispatch_rule: DispatchRule rule
    """
    new_report_id = ()
    for i in range(len(report_id)):
        if i >= len(primary_rule.fields):
            return new_report_id
        if dispatch_rule.fields[i] == primary_rule.fields[i]:
            new_report_id += (report_id[i], )
        else:
            return new_report_id
    return new_report_id


class DispatcherActor(Actor):
    """
    DispatcherActor class herited from Actor.

    Route message to the corresponding Formula, and create new one
    if no Formula exist for this message.
    """

    def __init__(self):
        Actor.__init__(self, DispatcherStartMessage)

        self.formula_class: Type[FormulaActor] = None
        self.formula_values: FormulaValues = None
        self.route_table: RouteTable = None
        self.device_id = None

        self._exit_mode = False
        self.formula_name_service = None
        self.formula_waiting_service = FormulaWaitingService()
        self.formula_pool = {}
        self.formula_number_id = 0

    def _initialization(self, message: StartMessage):
        Actor._initialization(self, message)
        self.formula_class = message.formula_class
        self.formula_values = message.formula_values
        self.route_table = message.route_table
        self.device_id = message.device_id

        self.formula_name_service = FormulaNameService()

        if self.route_table.primary_dispatch_rule is None:
            raise InitializationException('Dispatcher initialized without primary dispatch rule')

    def receiveMsg_PoisonMessage(self, message: PoisonMessage, sender: ActorAddress):
        """
        When receiving a PoisonMessage from a formula :
            - notify the formula blocking detector that a formula send a PoisonMessage
            - check from the blocking detector if the formula crashed
            - if the formula crashed, restart it
        """
        poison_message = message.poisonMessage
        for formula_name, (formula, blocking_detector) in self.formula_pool.items():
            if sender == formula:
                log_line = 'received poison messsage from formula ' + formula_name + ' for message ' + str(poison_message)
                log_line += 'with this error stack : ' + message.details
                self.log_debug(log_line)
                blocking_detector.notify_poison_received(poison_message)
                if blocking_detector.is_blocked():
                    self.log_debug('formula ' + formula_name + ' is blocked : ' + str(blocking_detector.is_blocked()))
                    self.log_debug('restart formula ' + formula_name)
                    self.log_error('formula ' + formula_name + ' is blocked after this error : ' + message.details)
                    self._restart_formula(formula_name)
                    return

    def receiveMsg_ActorExitRequest(self, message: ActorExitRequest, sender: ActorAddress):
        """
        When receiving ActorExitRequest, forward it to all formula
        """
        Actor.receiveMsg_ActorExitRequest(self, message, sender)
        for _, (formula, __) in self.formula_pool.items():
            self.send(formula, ActorExitRequest())
        for _, formula in self.formula_waiting_service.get_all_formula():
            self.send(formula, ActorExitRequest())

    def _gen_formula_name(self, formula_id):
        name = 'formula' + str(self.formula_number_id)
        self.formula_number_id += 1
        for field in formula_id:
            name += '__' + str(field)
        return name

    def _send_message(self, formula_name, message):
        try:
            formula, blocking_detector = self.formula_pool[formula_name]
            message.dispatcher_report_id = blocking_detector.get_message_id()
            self.log_debug('send ' + str(message) + ' to ' + formula_name)
            self.send(formula, message)
        except KeyError:
            self.formula_waiting_service.add_message(formula_name, message)

    def receiveMsg_Report(self, message: Report, _: ActorAddress):
        """
        When receiving a report, split it into sub-reports (if needed) and send them to their corresponding formula.
        If the corresponding formula does not exist, the dispatcher create it and send it the report
        """
        self.log_debug('received ' + str(message))
        dispatch_rule = self.route_table.get_dispatch_rule(message)
        primary_dispatch_rule = self.route_table.primary_dispatch_rule
        if dispatch_rule is None:
            self.log_warning('no dispatch rule for report ' + str(message))
            return
        formula_ids = _extract_formula_id(message, dispatch_rule, primary_dispatch_rule)

        for formula_id in formula_ids:
            primary_rule_fields = primary_dispatch_rule.fields
            if len(formula_id) == len(primary_rule_fields):
                try:
                    formula_name = self.formula_name_service.get_direct_formula_name(formula_id)
                    self._send_message(formula_name, message)
                except KeyError:
                    formula_name = self._gen_formula_name(formula_id)
                    self.log_info('create formula ' + formula_name)
                    formula = self._create_formula(formula_id, formula_name)
                    self.formula_name_service.add(formula_id, formula_name)
                    self.formula_waiting_service.add(formula_name, formula)
                    self.formula_waiting_service.add_message(formula_name, message)
            else:
                for formula_name in self.formula_name_service.get_corresponding_formula(list(formula_id)):
                    self._send_message(formula_name, message)

    def _get_formula_name_from_address(self, formula_address: ActorAddress):
        for name, (address, _) in self.formula_pool.items():
            if formula_address == address:
                return name
        return self.formula_waiting_service.get_formula_by_address(formula_address)

    def receiveMsg_ChildActorExited(self, message: ChildActorExited, _: ActorAddress):
        """
        When receive ChildActorExited from a formula:
        remove formula from formula pool and if dispatcher is in exit_mode and no formula is running, send an EndMessage to all pusher
        """
        try:
            formula_name = self._get_formula_name_from_address(message.childAddress)
        except AttributeError:
            return
        self.formula_name_service.remove_formula(formula_name)
        del self.formula_pool[formula_name]
        if self._exit_mode and not self.formula_pool:
            for _, pusher in self.formula_values.pushers.items():
                self.send(pusher, EndMessage(self.name))
            self.send(self.myAddress, ActorExitRequest())

    def receiveMsg_ErrorMessage(self, message: ErrorMessage, _: ActorAddress):
        """
        When receiving an ErrorMessage after trying to start a formula, remove formula from waiting service
        """
        self.log_info('error while trying to start ' + message.sender_name + ' : ' + message.error_message)
        self.formula_waiting_service.remove_formula(message.sender_name)

    def receiveMsg_OKMessage(self, message: OKMessage, sender: ActorAddress):
        """
        When receiving OKMessage after trying to start a formula, move formula from the waiting service to the formula pool
        """
        formula_name = message.sender_name
        waiting_messages = self.formula_waiting_service.get_waiting_messages(formula_name)
        self.formula_waiting_service.remove_formula(formula_name)
        self.formula_pool[formula_name] = (sender, BlockingDetector())
        for waiting_msg in waiting_messages:
            self._send_message(formula_name, waiting_msg)
        self.log_info('formula ' + formula_name + 'started')

    def receiveMsg_EndMessage(self, message: EndMessage, _: ActorAddress):
        """
        When receiving an EndMessage, set dispatcher into exit_mode and send an EndMessage to all formula
        """
        self.log_debug('received message ' + str(message))
        self._exit_mode = True
        for _, (formula, __) in self.formula_pool.items():
            self.send(formula, EndMessage(self.name))
        for formula_name, _ in self.formula_waiting_service.get_all_formula():
            self.formula_waiting_service.add_message(formula_name, message)

    def _restart_formula(self, formula_name: str):
        formula, _ = self.formula_pool[formula_name]
        formula_id = self.formula_name_service.get_formula_id(formula_name)

        # remove crashed formula
        self.formula_name_service.remove_formula(formula_name)
        del self.formula_pool[formula_name]
        self.send(formula, ActorExitRequest())

        # create new formula
        new_name = self._gen_formula_name(formula_id)

        formula = self._create_formula(formula_id, new_name)
        self.formula_name_service.add(formula_id, new_name)
        self.formula_waiting_service.add(new_name, formula)
        self.log_debug('restart formula' + formula_name + ' with new name : ' + new_name)

    def _create_formula(self, formula_id: Tuple, formula_name: str) -> ActorAddress:
        formula = self.createActor(self.formula_class)
        domain_values = self.formula_class.gen_domain_values(self.device_id, formula_id)
        start_message = FormulaStartMessage(self.name, formula_name, self.formula_values, domain_values)
        self.send(formula, start_message)
        return formula


class FormulaWaitingService:
    """
    Pool of fomula that received a StartMessage but didn't answer yet
    """
    def __init__(self):
        self.formulas = {}
        self.waiting_messages = {}

    def get_all_formula(self) -> List[Tuple[str, ActorAddress]]:
        """
        :return: list of tuple (formula_name, formula_address)
        """
        return self.formulas.items()

    def add(self, formula_name: str, formula_address: ActorAddress):
        """
        add a formula to the waiting service
        """
        self.formulas[formula_name] = formula_address
        self.waiting_messages[formula_name] = []

    def add_message(self, formula_name: str, message: Report):
        """
        store a message receive by a not started formula
        """
        self.waiting_messages[formula_name].append(message)

    def get_waiting_messages(self, formula_name: str) -> List[Report]:
        """
        :return: the list of message receive by the formula since the first StartMessage was send to the formula
        :raise AttributeError: if no formula with the given name exists
        """
        if formula_name in self.formulas:
            return self.waiting_messages[formula_name]
        raise AttributeError('unknow formula ' + str(formula_name))

    def get_formula_by_address(self, formula_address: ActorAddress) -> str:
        """
        :return: the formula name bind to the given formula address
        :raise AttributeError: if no formula with the given address exists
        """
        for name, address in self.formulas.items():
            if formula_address == address:
                return name
        raise AttributeError('no such formula with address ' + str(formula_address))

    def remove_formula(self, formula_name: str):
        """
        remove formula from the waiting service
        :raise AttributeError: if no formula with the given name exists
        """
        if formula_name in self.formulas:
            del self.formulas[formula_name]
            del self.waiting_messages[formula_name]
        else:
            raise AttributeError('unknow formula ' + str(formula_name))


class FormulaNameService:
    """
    Service that make correspondance between formula name and formula id
    """
    def __init__(self):
        self.formula_name = {}
        self.formula_tree = Tree()

    def add(self, formula_id, formula_name: str):
        """
        add a formula name into the service with its main id
        """
        self.formula_name[formula_id] = formula_name
        self.formula_tree.add(list(formula_id), formula_name)

    def get_direct_formula_name(self, formula_id) -> str:
        """
        Get the formula corresponding to the given formula id
        or create and return it if its didn't exist

        :param tuple formula_id: Key corresponding to a Formula
        :return: a Formula name
        """

        return self.formula_name[formula_id]

    def get_formula_id(self, formula_name_to_find: str) -> Tuple:
        """
        return main formula id from formula name
        """
        for formula_id, formula_name in self.formula_name.items():
            if formula_name == formula_name_to_find:
                return formula_id
        return None

    def get_corresponding_formula(self, formula_id):
        """
        Get the Formulas which have id match with the given formula_id

        :param tuple formula_id: Key corresponding to a Formula
        :return: All Formulas that match with the key
        :rtype: list(Formula)
        """
        return self.formula_tree.get(formula_id)

    def remove_formula(self, formula_name_to_remove: str):
        """
        remove from the pool the formula with the given address
        :param formula_address_to_remove: address of the formula to remove
        :raise AttributeError: if the pool doesn't contain any formula with this address
        """
        formula_id_to_remove = None
        for formula_id, formula_name in self.formula_name.items():
            if formula_name_to_remove == formula_name:
                formula_id_to_remove = formula_id
        if formula_id_to_remove is not None:
            del self.formula_name[formula_id_to_remove]
        else:
            raise AttributeError



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

from datetime import datetime
from typing import Dict, NewType, Tuple, List, Any
from powerapi.exception import PowerAPIExceptionWithMessage
from powerapi.message import Message


CSV_HEADER_COMMON = ['timestamp', 'sensor', 'target']
CsvLines = NewType('CsvLines', Tuple[List[str], Dict[str, str]])


class BadInputData(PowerAPIExceptionWithMessage):
    """
    Exception raised when input data can't be converted to a Report
    """
    def __init__(self, msg, input_data):
        PowerAPIExceptionWithMessage.__init__(self, msg)
        self.input_data = input_data


class Report(Message):
    """
    Report abtract class.
    """

    def __init__(self, timestamp: datetime, sensor: str, target: str, metadata: Dict[str, Any] = {}):
        """
        Initialize a report using the given parameters.
        :param datetime timestamp: Timestamp
        :param str sensor: Sensor name.
        :param str target: Target name.
        """
        Message.__init__(self, None)
        self.timestamp = timestamp
        self.sensor = sensor
        self.target = target
        self.metadata = dict(metadata)

        #: id given by the dispatcher actor in order manage report order
        self.dispatcher_report_id = None

    def __str__(self):
        return '%s(%s, %s, %s)' % (self.__class__.__name__, self.timestamp, self.sensor, self.target)

    def __repr__(self):
        return '%s(%s, %s, %s)' % (self.__class__.__name__, self.timestamp, self.sensor, self.target)

    def __eq__(self, other):
        return (isinstance(other, type(self)) and
                self.timestamp == other.timestamp and
                self.sensor == other.sensor and
                self.target == other.target and
                self.metadata == other.metadata)

    @staticmethod
    def to_json(report: Report) -> Dict:
        """
        :return: a json dictionary, that can be converted into json format, from a given Report
        """
        return report.__dict__

    @staticmethod
    def _extract_timestamp(ts):
        if isinstance(ts, str):
            try:
                return datetime.strptime(ts, "%Y-%m-%dT%H:%M:%S.%f")
            except ValueError:
                try:
                    return datetime.fromtimestamp(int(ts) / 1000)
                except ValueError:
                    ValueError('timestamp string have to be formated with the following format "%Y-%m-%dT%H:%M:%S.%f"')

        if isinstance(ts, datetime):
            return ts

        raise ValueError('timestamp must be a datetime.datetime or a string')



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

DEFAULT_BUFFER_SIZE = 4096


class JsonStream:
    """read data received from a input utf-8 byte stream socket as a json stream

    :param stream_reader:
    :param buffer_size: size of the buffer used to receive data from the socket,
                        it must match the average size of received json string
                        (default 4096 bytes)
    """

    def __init__(self, stream_reader, buffer_size=4096):
        self.stream_reader = stream_reader
        self.json_buffer = b''
        self.buffer_size = buffer_size
        self.open_brackets = 0

    async def _get_bytes(self):
        data = await self.stream_reader.read(n=self.buffer_size)
        return b'' if data is None else data

    def _extract_json_end_position(self, first_new_byte):
        """
        Find the first valable report in the stream
        """
        i = first_new_byte
        # print("buffer is ", self.json_buffer[first_new_byte:])
        if len(self.json_buffer) == 0:
            return -1
        if self.json_buffer[0] != 123:   # ASCII code opening bracket
            return -1

        while i < len(self.json_buffer):
            if self.json_buffer[i] == 125:  # ASCII code closing bracket
                self.open_brackets -= 1
                # print("opening : ", self.open_brackets)
            elif self.json_buffer[i] == 123:  # ASCII code opening bracket
                self.open_brackets += 1
                # print("closing :",self.open_brackets)
            if self.open_brackets == 0:
                return i
            i += 1

        return -1

    async def read_json_object(self):
        """
        return all the json object received from the connection as a iteration of string
        """
        if len(self.json_buffer) != 0 and self.open_brackets == 0:
            # Last iteration _extract_json_end_position returned a json_object
            # and breaked. If the buffer isn't empty wasn't treated, so we have
            # to treat it
            first_new_byte = 0
        else:
            first_new_byte = len(self.json_buffer)
            self.json_buffer += await self._get_bytes()
        i = self._extract_json_end_position(first_new_byte)

        if i == -1:
            return None
        if i == len(self.json_buffer) - 1:
            json_str = self.json_buffer[:]
            self.json_buffer = b''
            # print("buffer empty")
            # print(self.open_brackets)
        else:
            json_str = self.json_buffer[:i + 1]
            self.json_buffer = self.json_buffer[i + 1:]
            # print("buffer non empty ")
            # print(self.open_brackets)
        return json_str.decode('utf-8')



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import logging
from typing import List, Type
try:
    from prometheus_client import start_http_server, Gauge
except ImportError:
    logging.getLogger().info("prometheus-client is not installed.")

from powerapi.report import Report
from powerapi.utils import StatBuffer
from .base_db import BaseDB


class BasePrometheusDB(BaseDB):
    """
    Base class to expose data to prometheus instance
    """
    def __init__(self, report_type: Type[Report], port: int, address: str, metric_name: str,
                 metric_description: str, tags: List[str]):
        BaseDB.__init__(self, report_type)
        self.address = address
        self.port = port
        self.metric_name = metric_name
        self.metric_description = metric_description
        self.tags = tags

    def _init_metrics(self):
        raise NotImplementedError()

    def connect(self):
        """
        Start a HTTP server exposing metrics
        """
        self._init_metrics()
        start_http_server(self.port)


class PrometheusDB(BasePrometheusDB):
    """
    Database that expose received data as metric in order to be scrapped by a prometheus instance
    Could only be used with a pusher actor
    """

    def __init__(self, report_type: Type[Report], port: int, address: str, metric_name: str,
                 metric_description: str, aggregation_periode: int, tags: List[str]):
        """
        :param address:             address that expose the metric
        :param port:
        :param metric_name:
        :param metric_description:  short sentence that describe the metric
        :param aggregation_periode: number of second for the value must be aggregated before compute statistics on them
        :param tags: metadata used to tag metric
        """
        BasePrometheusDB.__init__(self, report_type, port, address, metric_name, metric_description, tags)
        self.aggregation_periode = aggregation_periode
        self.final_tags = ['sensor', 'target'] + tags

        self.mean_metric = None
        self.std_metric = None
        self.min_metric = None
        self.max_metric = None

        self.exposed_measure = {}
        self.measure_for_current_period = {}
        self.current_period_end = 0

        self.buffer = StatBuffer(aggregation_periode)

    def __iter__(self):
        raise NotImplementedError()

    def _init_metrics(self):
        self.mean_metric = Gauge(self.metric_name + '_mean', self.metric_description + '(MEAN)', self.final_tags)
        self.std_metric = Gauge(self.metric_name + '_std', self.metric_description + '(STD)', self.final_tags)
        self.min_metric = Gauge(self.metric_name + '_min', self.metric_description + '(MIN)', self.final_tags)
        self.max_metric = Gauge(self.metric_name + '_max', self.metric_description + '(MAX)', self.final_tags)

    def _expose_data(self, key):
        aggregated_value = self.buffer.get_stats(key)
        if aggregated_value is None:
            return

        kwargs = {label: aggregated_value['tags'][label] for label in self.final_tags}
        try:
            self.mean_metric.labels(**kwargs).set(aggregated_value['mean'])
            self.std_metric.labels(**kwargs).set(aggregated_value['std'])
            self.min_metric.labels(**kwargs).set(aggregated_value['min'])
            self.max_metric.labels(**kwargs).set(aggregated_value['max'])
        except TypeError:
            self.mean_metric.labels(kwargs).set(aggregated_value['mean'])
            self.std_metric.labels(kwargs).set(aggregated_value['std'])
            self.min_metric.labels(kwargs).set(aggregated_value['min'])
            self.max_metric.labels(kwargs).set(aggregated_value['max'])

    def _report_to_measure_and_key(self, report):
        value = self.report_type.to_prometheus(report, self.tags)
        key = ''.join([str(value['tags'][tag]) for tag in self.final_tags])
        return key, value

    def _update_exposed_measure(self):
        updated_exposed_measure = {}

        for key in self.exposed_measure:
            if key not in self.measure_for_current_period:
                args = self.exposed_measure[key]
                self.mean_metric.remove(*args)
                self.std_metric.remove(*args)
                self.min_metric.remove(*args)
                self.max_metric.remove(*args)
            else:
                updated_exposed_measure[key] = self.exposed_measure[key]
        self.exposed_measure = updated_exposed_measure

    def _append_measure_from_old_period_to_buffer_and_expose_data(self):
        for old_key, old_measure_list in self.measure_for_current_period.items():
            for old_measure in old_measure_list:
                self.buffer.append(old_measure, old_key)
            self._expose_data(old_key)

    def _reinit_persiod(self, new_measure_time):
        self.current_period_end = new_measure_time + self.aggregation_periode
        self.measure_for_current_period = {}

    def save(self, report: Report):
        """
        Override from BaseDB

        :param report: Report to save
        """
        key, measure = self._report_to_measure_and_key(report)
        if measure['time'] > self.current_period_end:
            self._append_measure_from_old_period_to_buffer_and_expose_data()
            self._update_exposed_measure()
            self._reinit_persiod(measure['time'])

        if key not in self.exposed_measure:
            args = [measure['tags'][label] for label in self.final_tags]
            self.exposed_measure[key] = args

        if key not in self.measure_for_current_period:
            self.measure_for_current_period[key] = []

        self.measure_for_current_period[key].append(measure)

    def save_many(self, reports: List[Report]):
        """
        Save a batch of data

        :param reports: Batch of data.
        """
        for report in reports:
            self.save(report)



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import Dict, List, Tuple
import numpy as np


def _compute_stats(values: List[float]) -> Tuple[float, float, float, float]:
    """
    compute mean, std, min, max from a list of float
    """
    np_values = np.array([value['value'] for value in values])
    return np_values.mean(), np_values.std(), np_values.min(), np_values.max()


class StatBuffer:
    """
    Buffer that store timeseries values and compute statistics on it
    """

    def __init__(self, aggregation_periode: int):
        """
        :param aggregation_periode: number of second for the value must be aggregated before compute statistics on them
        """

        self.aggregation_periode = aggregation_periode
        self.buffer = {}

    def append(self, measure: Dict, key: str):
        """
        append a measure to the buffer

        measure must have the following format :

        {
          'tags': Dict,
          'time': int,
          'value': float,
        }
        """
        if key not in self.buffer:
            self.buffer[key] = []

        self.buffer[key].append(measure)

    def is_available(self, key: str) -> bool:
        """
        Return true if they are enough value for the given key to compute a statistics (depend on aggregation periode parameter)
        :raise KeyError: if no measure were append with this key before
        """
        if key not in self.buffer:
            raise KeyError

        first_measure = self.buffer[key][0]
        last_measure = self.buffer[key][-1]

        return (last_measure['time'] - first_measure['time']) >= self.aggregation_periode

    def _split_values(self, values: List[Dict]):
        time_of_first_measure = values[0]['time']

        def split(value_in_periode, value_out_periode):
            if value_out_periode == []:
                return value_in_periode, value_out_periode
            if value_out_periode[0]['time'] - time_of_first_measure > self.aggregation_periode:
                return value_in_periode, value_out_periode
            else:
                val = value_out_periode.pop(0)
                value_in_periode.append(val)
                return split(value_in_periode, value_out_periode)

        return split([], values)

    def get_stats(self, key: str) -> Dict:
        """
        return statistics on the values corresponding to the given key

        :return: a dictionary with this format:
        :raise KeyError: if no measure were append with this key before
        {
          'mean': float,
          'std': float,
          'min': float,
          'max': float',
          'time': int, # timestamp of the last measure
          'tags': Dict,
        }
        return None if no statistics are available
        """
        if not self.is_available(key):
            return None

        values, self.buffer[key] = self._split_values(self.buffer[key])

        mean, std, min, max = _compute_stats(values)

        return {
            'mean': mean,
            'std': std,
            'min': min,
            'max': max,
            'time': values[-1]['time'],
            'tags': values[-1]['tags']
        }



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from powerapi.exception import PowerAPIException


class PrimaryDispatchRuleRuleAlreadyDefinedException(PowerAPIException):
    """
    Exception raised when trying to define a primary dispatch rule on a route table that have already one
    """


class RouteTable:
    """
    Structure that map a :class:`Report<powerapi.report.Report>` type to a
    :class:`DispatchRule<powerapi.dispatch_rule.DispatchRule>` rule
    """

    def __init__(self):
        #: (array): Array of tuple that link a Report type to a DispatchRule
        # rule
        self.route_table = []
        #: (powerapi.DispatchRule): Allow to define how to create the Formula id
        self.primary_dispatch_rule = None

    def get_dispatch_rule(self, msg):
        """
        Return the corresponding group by rule mapped to the received message
        type

        :param type msg: the received message
        :return: the dispatch_rule rule mapped to the received message type
        :rtype: powerapi.dispatch_rule.DispatchRule
        :raise: UnknowMessageTypeException if no group by rule is mapped to the
                received message type
        """
        for (report_class, dispatch_rule) in self.route_table:
            if isinstance(msg, report_class):
                return dispatch_rule

        return None

    def dispatch_rule(self, report_class, dispatch_rule):
        """
        Add a dispatch_rule rule to the route table

        :param Type report_class: Type of the message that the
                                  dispatch_rule rule must handle
        :param dispatch_rule: Group_by rule to add
        :type dispatch_rule:  powerapi.dispatch_rule.DispatchRule
        """
        if dispatch_rule.is_primary:
            if self.primary_dispatch_rule is not None:
                raise PrimaryDispatchRuleRuleAlreadyDefinedException()
            self.primary_dispatch_rule = dispatch_rule

        self.route_table.append((report_class, dispatch_rule))



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import json
from typing import Dict, List

from powerapi.report import HWPCReport
import powerapi.test_utils.report as parent_module
from powerapi.test_utils.libvirt import LIBVIRT_TARGET_NAME1, LIBVIRT_TARGET_NAME2


###################
# Report Creation #
###################
def extract_rapl_reports_with_2_sockets(number_of_reports: int) -> List[Dict]:
    """
    Extract the number_of_reports first reports of the file hwpc_rapl_2_socket.json
    This file contain hwpc reports with only RAPL_PKG events, recorded on a two socket host
    :return: a list of reports. Each reports is a json dictionary
    """
    path = parent_module.__path__[0]
    json_file = open(path + '/hwpc_rapl_2_socket.json', 'r')
    reports = json.load(json_file)
    json_file.close()
    return reports['reports'][:number_of_reports]


def extract_all_events_reports_with_2_sockets(number_of_reports: int) -> List[Dict]:
    """
    Extract the number_of_reports first reports of the file hwpc_rapl_2_socket.json
    This file contain hwpc reports , recorded on a two socket host, with events :
    - RAPL_PKG
    - MPERF
    - APERF
    - TSC
    - CPU_CLK_THREAD_UNHALTED:THREAD_P
    - CPU_CLK_THREAD_UNHALTED:REF_P
    - LLC_MISSES
    - INSTRUCTIONS_RETIRED
    :return: a list of reports. Each reports is a json dictionary
    """
    path = parent_module.__path__[0]
    json_file = open(path + '/hwpc_all_events_2_socket.json', 'r')
    reports = json.load(json_file)
    json_file.close()
    return reports['reports'][:number_of_reports]


def extract_all_events_reports_with_vm_name(number_of_reports: int) -> List[Dict]:
    """
    Extract the number_of_reports first reports of the file hwpc_all_vm_target.json

    This file contain hwpc reports , recorded on a two socket host, with events :
    - RAPL_PKG
    - MPERF
    - APERF
    - TSC
    - CPU_CLK_THREAD_UNHALTED:THREAD_P
    - CPU_CLK_THREAD_UNHALTED:REF_P
    - LLC_MISSES
    - INSTRUCTIONS_RETIRED

    ! only reports with target LIBVIRT_TARGET_NAME1 or LIBVIRT_TARGET_NAME2 are returned !

    :return: a list of reports. Each reports is a json dictionary. Only reports with vm name as target is returned
    """
    path = parent_module.__path__[0]
    json_file = open(path + '/hwpc_all_vm_target.json', 'r')
    reports = json.load(json_file)
    json_file.close()
    return list(filter(lambda r: r['target'] == LIBVIRT_TARGET_NAME1 or r['target'] == LIBVIRT_TARGET_NAME2, reports['reports']))[:number_of_reports]


def gen_HWPCReports(number_of_reports: int) -> List[HWPCReport]:
    """
    generate number_of_reports HWPCReport extracted from the file hwpc.json
    :return: a list of HWPCReport
    """
    path = parent_module.__path__[0]
    json_file = open(path + '/hwpc.json', 'r')
    reports = list(map(HWPCReport.from_json, json.load(json_file)['reports']))

    json_file.close()
    return reports[:number_of_reports]



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import List, Type
from powerapi.report import Report
from powerapi.exception import PowerAPIExceptionWithMessage


class DBError(PowerAPIExceptionWithMessage):

    """
    Error raised when an error occuried when using a database
    """
    def __init__(self, msg: str):
        PowerAPIExceptionWithMessage.__init__(self, msg)


class IterDB:
    """
    IterDB class

    This class allows to browse a database as an iterable
    """

    def __init__(self, db, report_type, stream_mode):
        """
        """
        self.db = db
        self.stream_mode = stream_mode
        self.report_type = report_type

    def __iter__(self):
        """
        """
        raise NotImplementedError()

    def __next__(self) -> Report:
        """
        """
        raise NotImplementedError()


class BaseDB:
    """
    Abstract class which define every common function for database uses.

    This class define every common function that need to be implemented
    by each DB module. A database module correspond to a kind of BDD.
    For example, Mongodb, influxdb, csv are different kind of BDD.
    """
    def __init__(self, report_type: Type[Report]):
        self.asynchrone = False
        self.exceptions = []
        self.report_type = report_type

    def connect(self):
        """
        Function that allow to load the database. Depending of the type,
        different process can happen.

        .. note:: Need to be overrided
        """
        raise NotImplementedError()

    def iter(self, stream_mode: bool) -> IterDB:
        """
        Create the iterator for get the data
        :param stream_mode: Define if we read in stream mode
        """
        raise NotImplementedError()

    def save(self, report: Report):
        """
        Allow to save a json input in the db

        :param report: Report
        """
        raise NotImplementedError()

    def save_many(self, reports: List[Report]):
        """
        Allow to save a batch of data

        :param reports: Batch of Serialized Report
        """
        raise NotImplementedError()



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import getopt
import sys
from copy import deepcopy

from powerapi.exception import PowerAPIException


def _extract_minus(arg):
    if len(arg) > 2:
        return arg[2:]
    return arg[1]


def _cast_argument_value(arg_name, val, action):
    if not action.is_flag:
        try:
            return action.arg_type(val)
        except ValueError as exn:
            raise BadTypeException(arg_name, action.arg_type) from exn
    return val


def _find_longest_name(names):
    max_len = 0
    longest_name = ''
    for name in names:
        if len(name) > max_len:
            longest_name = name
            max_len = len(name)
    return longest_name


#############
# EXCEPTION #
#############
class ParserException(PowerAPIException):
    """
    Base Exception for parser error
    """
    def __init__(self, argument_name):
        PowerAPIException.__init__(self)
        self.argument_name = argument_name


class NoNameSpecifiedForComponentException(ParserException):
    """
    Exception raised when attempting to parse substring thant describe a component which not contains the component name
    """


class ComponentAlreadyExistException(ParserException):
    """
    Exception raised when attempting to parse a substring to create a component with a name that already exist
    """


class SubParserWithoutNameArgumentException(PowerAPIException):
    """
    Exception raised when a subparser without argument name is added to a parser
    """


class TooManyArgumentNamesException(ParserException):
    """
    Exception raised when attemtping to add an argument with too much names

    """
    def __init__(self, argument_name):
        ParserException.__init__(self, argument_name)


class AlreadyAddedArgumentException(ParserException):
    """
    Exception raised when attempting to add an argument to a parser that already
    have this argument

    """
    def __init__(self, argument_name):
        ParserException.__init__(self, argument_name)
        self.msg = 'Parser already contain an argument ' + argument_name


class AlreadyAddedSubparserException(ParserException):
    """
    Exception raised when attempting to add an argument to a parser that already
    have this argument

    """
    def __init__(self, argument_name):
        ParserException.__init__(self, argument_name)
        self.msg = 'Parser already contain SubParser with name ' + argument_name


class MissingArgumentException(ParserException):
    """
    Exception raised when a mandatory argument is missing
    """

    def __init__(self, argument_name):
        ParserException.__init__(self, argument_name)
        self.msg = 'Argument ' + argument_name + ' is missing'


class MissingValueException(ParserException):
    """
    Exception raised when an argument that require a value is caught without
    its value

    """
    def __init__(self, argument_name):
        ParserException.__init__(self, argument_name)
        self.msg = 'Argument ' + argument_name + ' require a value'


class UnknowArgException(ParserException):
    """
    Exception raised when the parser catch an argument that it can't handle

    """
    def __init__(self, argument_name):
        ParserException.__init__(self, argument_name)
        self.msg = 'Unknow argument ' + argument_name


class BadTypeException(ParserException):
    """
    Exception raised when an argument is parsed with a value of an incorrect type
    """
    def __init__(self, argument_name, arg_type):
        ParserException.__init__(self, argument_name)
        self.msg = argument_name + " expect " + arg_type.__name__


class BadContextException(ParserException):
    """
    Exception raised when the parser catch an argument that it can't handle in
    the current context
    """
    def __init__(self, argument_name, context_list):
        ParserException.__init__(self, argument_name)
        self.context_list = context_list
        self.msg = 'argument ' + argument_name + 'not used in the correct context\nUse it with the following arguments :'
        for main_arg_name, context_name in context_list:
            self.msg += '\n  --' + main_arg_name + ' ' + context_name


#################
# ACTION LAMBDA #
#################
def store_val(arg, val, args, acc):
    """
    lambda that store the value of the argument on the parser result

    """
    if val == '':
        val = None

    acc[arg] = val
    return args, acc


def store_true(arg, _, args, acc):
    """
    lambda that store a True boolean value on the parser result

    """
    acc[arg] = True
    return args, acc


class ParserAction:
    """
    Action binded to an argument
    """
    def __init__(self, name_list, is_flag, action, default_value, help_str, arg_type):
        self.name_list = name_list
        self.is_flag = is_flag
        self.action = action
        self.default_value = default_value
        self.help_str = help_str
        self.arg_type = arg_type


class SubParserGroup:
    """"""
    def __init__(self, group_name, help_str=''):
        self.group_name = group_name
        self.help_str = help_str
        self.subparsers = {}

    def contains(self, name):
        """"""
        return name in self.subparsers

    def add_subparser(self, name, subparser):
        """"""
        self.subparsers[name] = subparser

    def get_subparser(self, name):
        """"""
        return self.subparsers[name]

    def __iter__(self):
        return iter(self.subparsers.items())

    def get_help(self):
        """
        return help string
        """
        s = self.group_name + ' details :\n'
        for subparser_name, subparser in self.subparsers.items():
            s += '  --' + self.group_name + ' ' + subparser_name + ':\n'
            s += subparser.get_help()
            s += '\n'

        return s


class Parser:
    """"""
    def __init__(self):
        self.actions = {}
        self.default_values = {}
        self.action_list = []

    def add_argument(self, *names, flag=False, action=store_val, default=None,
                     help='', type=str):
        """add an optional argument to the parser that will activate an action

        :param str *names: names of the optional argument that will be bind to
                           the action (could be long or short name)

        :param bool flag: True if the argument doesn't require to be followed
                           by a value

        :param lambda action: action that will be executed when the argument is
                              caught by the parser. the lambda take 4 parameters
                              (the name of the argument caught by the parser,
                              the value attached to this arguemnt, the current
                              list of arguments that is parsed by the parser and
                              the parser result) and return a list of token and
                              a parser result(dict)

        :param default: the default value attached to this argument

        :param str help: string that describe the argument

        :param type type: type of the value that the argument must catch

        :raise AlreadyAddedArgumentException: when attempting to add an
                                               argument that already have been
                                               added to this parser

        """
        parser_action = ParserAction(list(names), flag, action, default, help, type)

        for name in names:
            if name in self.actions:
                raise AlreadyAddedArgumentException(name)
            self.actions[name] = parser_action

        action_name = _find_longest_name(names)

        if default is not None:
            self.default_values[action_name] = default

        self.action_list.append(parser_action)

    def _get_action_list_str(self, indent):
        s = ''
        for action in self.action_list:
            s += indent + ', '.join(map(lambda x: '-' + x if len(x) == 1 else '--' + x, action.name_list))
            s += ' : ' + action.help_str + '\n'
        return s

    def _unknow_argument_behaviour(self, arg_name, val, args, acc):
        raise NotImplementedError()

    def _parse(self, args, acc):
        while args != []:
            arg, val = args.pop(0)

            if arg not in self.actions:
                args.insert(0, (arg, val))
                return self._unknow_argument_behaviour(arg, val, args, acc)

            action = self.actions[arg]

            arg_long_name = _find_longest_name(action.name_list)
            val = _cast_argument_value(arg_long_name, val, action)

            args, acc = action.action(arg_long_name, val, args, acc)

        return args, acc


class ComponentSubParser(Parser):
    """"""
    def __init__(self, name):
        Parser.__init__(self)
        self.name = name

    def _unknow_argument_behaviour(self, arg_name, val, args, acc):
        return args, acc

    def subparse(self, token_list):
        """
        Parse the given token list until an unknow argument is caught

        :param list token_list: the token list currently parsed

        :result: the current result of the parser

        :return (list, dict): a tuple containing the token list without the
                              parsed argument and the result of the parsing

        """
        local_result = deepcopy(self.default_values)
        if token_list == []:
            return token_list, local_result

        return self._parse(token_list, local_result)

    def get_help(self):
        """
        return help string
        """
        return self._get_action_list_str('    ')


class MainParser(Parser):
    """"""
    def __init__(self, help_arg=True):
        """
        :param bool help_arg: if True, add a -h/--help argument that display help
        """
        Parser.__init__(self)
        self.short_arg = ''
        self.long_arg = []
        self.subparsers_group = {}

        self.help_arg = help_arg
        if help_arg:
            self.add_argument('h', 'help', flag=True, action=None)

    def get_help(self):
        """
        return help string
        """
        s = 'main arguments:\n'
        s += self._get_action_list_str('  ')
        s += '\n'

        for _, subparser_group in self.subparsers_group.items():
            s += subparser_group.get_help()

        return s

    def parse(self, args):
        """
        :param str args: string that contains the arguments and their values

        :return dict:

        :raise UnknowArgException: when the parser catch catch an argument that
                                   this parser can't handle

        :raise BadContextException: when an argument that the parser can't
                                    handle in the current context is caught

        :raise MissingValueException: when an argument that require a value is
                                      caught without its value

        :raise BadTypeException: when an argument is parsed with a value of an
                                 incorrect type
        """
        try:
            args, _ = getopt.getopt(args, self.short_arg, self.long_arg)
        except getopt.GetoptError as exn:
            if 'recognized' in exn.msg:
                raise UnknowArgException(exn.opt) from exn
            elif 'requires' in exn.msg:
                raise MissingValueException(exn.opt) from exn

        # retirer les moins
        args = list(map(lambda x: (_extract_minus(x[0]), x[1]), args))

        # verify if help argument exists in args

        if self.help_arg:
            for arg_name, _ in args:
                if arg_name in ('h', 'help'):
                    print(self.get_help())
                    sys.exit(0)

        acc = deepcopy(self.default_values)

        args, acc = self._parse(args, acc)

        return acc

    def _unknow_argument_behaviour(self, arg_name, val, args, acc):
        good_contexts = []
        for main_arg_name, subparser_group in self.subparsers_group.items():
            for subparser_name, subparser in subparser_group:
                if arg_name in subparser.actions:
                    good_contexts.append((main_arg_name, subparser_name))
        raise BadContextException(arg_name, good_contexts)

    def _add_argument_names(self, names, is_flag):

        if len(names) > 2:
            raise TooManyArgumentNamesException(names[2])

        if len(names) > 1 and len(names[0]) == len(names[1]):
            raise TooManyArgumentNamesException(names[1])

        def gen_name(name):
            if len(name) == 1:
                return name + ('' if is_flag else ':')
            return name + ('' if is_flag else '=')

        for name in names:
            if len(name) == 1:
                self.short_arg += gen_name(name)
            else:
                self.long_arg.append(gen_name(name))

    def add_argument(self, *names, flag=False, action=store_val, default=None, help='', type=str):
        Parser.add_argument(self, *names, flag=flag, action=action, default=default, help=help, type=type)
        self._add_argument_names(names, flag)

    def add_component_subparser(self, component_type, subparser, help_str=''):
        """
        Add a subparser that will be used by the argument *component_name*

        :param str component_type:
        :param ComponentSubParser subparser:

        :raise AlreadyAddedArgumentException: when attempting to add an
                                              argument that already have been
                                              added to this parser
        """
        def _action(arg, val, args, acc):
            if arg not in acc:
                acc[arg] = {}

            subparser = self.subparsers_group[arg].get_subparser(val)
            args, subparse_result = subparser.subparse(args)

            acc[arg][subparser.name] = subparse_result
            # acc[arg] = subparse_result
            return args, acc

        if component_type not in self.subparsers_group:
            self.subparsers_group[component_type] = SubParserGroup(component_type, help_str=help_str)
            self.add_argument(component_type, action=_action, help=help_str)
        else:
            if self.subparsers_group[component_type].contains(subparser.name):
                raise AlreadyAddedArgumentException(subparser.name)

        self.subparsers_group[component_type].add_subparser(subparser.name, subparser)

        for action_name, action in subparser.actions.items():
            self._add_argument_names([action_name], action.is_flag)

    def add_actor_subparser(self, component_type, subparser, help_str=''):
        """
        Add a subparser that will be used by the argument *component_name*
        The component must contain a name action
        :param str component_type:
        :param ComponentSubParser subparser:

        :raise AlreadyAddedArgumentException: when attempting to add an
                                              argument that already have been
                                              added to this parser
        """
        def _action(arg, val, args, acc):
            if arg not in acc:
                acc[arg] = {}

            subparser = self.subparsers_group[arg].get_subparser(val)
            args, subparse_result = subparser.subparse(args)

            if 'name' not in subparse_result:
                raise NoNameSpecifiedForComponentException(component_type)

            component_name = subparse_result['name']
            del subparse_result['name']

            if component_name in acc[arg]:
                raise ComponentAlreadyExistException(component_name)

            acc[arg][component_name] = subparse_result
            acc[arg][component_name]['type'] = subparser.name

            return args, acc

        if 'name' not in subparser.actions:
            raise SubParserWithoutNameArgumentException()
        if component_type not in self.subparsers_group:
            self.subparsers_group[component_type] = SubParserGroup(component_type, help_str=help_str)
            self.add_argument(component_type, action=_action, help=help_str)
        else:
            if self.subparsers_group[component_type].contains(subparser.name):
                raise AlreadyAddedArgumentException(subparser.name)

        self.subparsers_group[component_type].add_subparser(subparser.name, subparser)

        for action_name, action in subparser.actions.items():
            self._add_argument_names([action_name], action.is_flag)



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import sys
from typing import Dict, Tuple, Type

from powerapi.actor import Actor
from powerapi.exception import PowerAPIException
from powerapi.report import HWPCReport, PowerReport, ControlReport, ProcfsReport
from powerapi.database import MongoDB, CsvDB, InfluxDB, OpenTSDB, SocketDB, PrometheusDB, DirectPrometheusDB, VirtioFSDB, FileDB
from powerapi.puller import PullerActor
from powerapi.pusher import PusherActor
from powerapi.message import StartMessage, PusherStartMessage, PullerStartMessage
from powerapi.report_modifier.libvirt_mapper import LibvirtMapper


class Generator:
    """
    Generate an actor class and actor start message from config dict
    """

    def __init__(self, component_group_name):
        self.component_group_name = component_group_name

    def generate(self, config: Dict) -> Dict[str, Tuple[Type[Actor], StartMessage]]:
        """
        Generate an actor class and actor start message from config dict
        """
        if self.component_group_name not in config:
            print('Configuration error : no ' + self.component_group_name + ' specified', file=sys.stderr)
            raise PowerAPIException('Configuration error : no ' + self.component_group_name + ' specified')

        actors = {}

        for component_name, component_config in config[self.component_group_name].items():
            try:
                component_type = component_config['type']
                actors[component_name] = self._gen_actor(component_type, component_config, config, component_name)
            except KeyError as exn:
                msg = 'Configuration error : argument ' + exn.args[0]
                msg += ' needed with output ' + component_type
                print(msg, file=sys.stderr)
                raise PowerAPIException(msg) from exn

        return actors

    def _gen_actor(self, component_type: str, component_config: Dict, main_config: Dict, component_name: str) -> Tuple[Type[Actor], StartMessage]:
        raise NotImplementedError()


class ModelNameAlreadyUsed(PowerAPIException):
    """
    Exception raised when attempting to add to a DBActorGenerator a model factory with a name already bound to another
    model factory in the DBActorGenerator
    """
    def __init__(self, model_name):
        PowerAPIException.__init__(self)
        self.model_name = model_name


class DatabaseNameAlreadyUsed(PowerAPIException):
    """
    Exception raised when attempting to add to a DBActorGenerator a database factory with a name already bound to another
    database factory in the DBActorGenerator
    """
    def __init__(self, database_name):
        PowerAPIException.__init__(self)
        self.database_name = database_name


class ModelNameDoesNotExist(PowerAPIException):
    """
    Exception raised when attempting to remove to a DBActorGenerator a model factory with a name that is not bound to another
    model factory in the DBActorGenerator
    """
    def __init__(self, model_name):
        PowerAPIException.__init__(self)
        self.model_name = model_name


class DatabaseNameDoesNotExist(PowerAPIException):
    """
    Exception raised when attempting to remove to a DBActorGenerator a database factory with a name that is not bound to another
    database factory in the DBActorGenerator
    """
    def __init__(self, database_name):
        PowerAPIException.__init__(self)
        self.database_name = database_name


def gen_tag_list(db_config: Dict):
    """
    Generate tag list from tag string
    """
    if 'tags' not in db_config:
        return []
    return db_config['tags'].split(',')


class DBActorGenerator(Generator):
    """
    ActorGenerator that initialise the start message with a database from config
    """
    def __init__(self, component_group_name):
        Generator.__init__(self, component_group_name)
        self.model_factory = {
            'HWPCReport': HWPCReport,
            'PowerReport': PowerReport,
            'ControlReport': ControlReport,
            'ProcfsReport': ProcfsReport
        }

        self.db_factory = {
            'mongodb': lambda db_config: MongoDB(db_config['model'], db_config['uri'], db_config['db'], db_config['collection']),
            'socket': lambda db_config: SocketDB(db_config['model'], db_config['port']),
            'csv': lambda db_config: CsvDB(db_config['model'], gen_tag_list(db_config),
                                           current_path=os.getcwd() if 'directory' not in db_config else db_config['directory'],
                                           files=[] if 'files' not in db_config else db_config['files']),
            'influxdb': lambda db_config: InfluxDB(db_config['model'], db_config['uri'], db_config['port'], db_config['db'], gen_tag_list(db_config)),
            'opentsdb': lambda db_config: OpenTSDB(db_config['model'], db_config['uri'], db_config['port'], db_config['metric_name']),
            'prom': lambda db_config: PrometheusDB(db_config['model'], db_config['port'], db_config['uri'], db_config['metric_name'],
                                                   db_config['metric_description'], db_config['aggregation_period'], gen_tag_list(db_config)),
            'direct_prom': lambda db_config: DirectPrometheusDB(db_config['model'], db_config['port'], db_config['uri'], db_config['metric_name'],
                                                                db_config['metric_description'], gen_tag_list(db_config)),
            'virtiofs': lambda db_config: VirtioFSDB(db_config['model'], db_config['vm_name_regexp'], db_config['root_directory_name'],
                                                     db_config['vm_directory_name_prefix'], db_config['vm_directory_name_suffix']),
            'filedb': lambda db_config: FileDB(db_config['model'], db_config['filename'])
        }

    def remove_model_factory(self, model_name):
        """
        remove a Model from generator
        """
        if model_name not in self.model_factory:
            raise ModelNameDoesNotExist(model_name)
        del self.model_factory[model_name]

    def remove_db_factory(self, database_name):
        """
        remove a database from generator
        """
        if database_name not in self.db_factory:
            raise DatabaseNameDoesNotExist(database_name)
        del self.db_factory[database_name]

    def add_model_factory(self, model_name, model_factory):
        """
        add a model to generator
        """
        if model_name in self.model_factory:
            raise ModelNameAlreadyUsed(model_name)
        self.model_factory[model_name] = model_factory

    def add_db_factory(self, db_name, db_factory):
        """
        add a database to generator
        """
        if db_name in self.db_factory:
            raise DatabaseNameAlreadyUsed(db_name)
        self.db_factory[db_name] = db_factory

    def _generate_db(self, db_name, db_config, _):
        if db_name not in self.db_factory:
            msg = 'Configuration error : database type ' + db_name + ' unknow'
            print(msg, file=sys.stderr)
            raise PowerAPIException(msg)
        else:
            return self.db_factory[db_name](db_config)

    def _generate_model(self, model_name, db_config):
        if model_name not in self.model_factory:
            msg = 'Configuration error : model type ' + model_name + ' unknow'
            print(msg, file=sys.stderr)
            raise PowerAPIException(msg)
        else:
            return self.model_factory[db_config['model']]

    def _gen_actor(self, db_name, db_config, main_config, actor_name):
        model = self._generate_model(db_config['model'], db_config)
        db_config['model'] = model
        db = self._generate_db(db_name, db_config, main_config)
        start_message = self._start_message_factory(actor_name, db, model, main_config['stream'], main_config['verbose'])
        actor = self._actor_factory(db_config)
        return (actor, start_message)

    def _actor_factory(self, db_config):
        raise NotImplementedError()

    def _start_message_factory(self, name, db, model, stream_mode, level_logger):
        raise NotImplementedError()


class PullerGenerator(DBActorGenerator):
    """
    Generate Puller Actor class and Puller start message from config
    """
    def __init__(self, report_filter, report_modifier_list=[]):
        DBActorGenerator.__init__(self, 'input')
        self.report_filter = report_filter
        self.report_modifier_list = report_modifier_list

    def _actor_factory(self, db_config):
        return PullerActor

    def _start_message_factory(self, name, db, model, stream_mode, level_logger):
        return PullerStartMessage('system', name, db, self.report_filter, stream_mode, report_modifiers=self.report_modifier_list)


class PusherGenerator(DBActorGenerator):
    """
    Generate Pusher actor and Pusher start message from config
    """

    def __init__(self):
        DBActorGenerator.__init__(self, 'output')

    def _actor_factory(self, db_config):
        return PusherActor

    def _start_message_factory(self, name, db, model, stream_mode, level_logger):
        return PusherStartMessage('system', name, db)


class ReportModifierGenerator:
    """
    Generate Report modifier list from config
    """
    def __init__(self):
        self.factory = {'libvirt_mapper': lambda config: LibvirtMapper(config['uri'], config['domain_regexp'])}

    def generate(self, config: Dict):
        """
        Generate Report modifier list from config
        """
        report_modifier_list = []
        if 'report_modifier' not in config:
            return []
        for report_modifier_name in config['report_modifier'].keys():
            report_modifier = self.factory[report_modifier_name](config['report_modifier'][report_modifier_name])
            report_modifier_list.append(report_modifier)
        return report_modifier_list



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import datetime
import logging
import sys
from powerapi.exception import PowerAPIException


class WrongFormatReport(PowerAPIException):
    """
    Exception raised when Sync receive a report which is not of a handled type
    """
    def __init__(self, report_type):
        PowerAPIException.__init__(self)
        self.report_type = report_type


class WrongTypeParameter(PowerAPIException):
    """
    Exception raised when Sync is instantiate with a parameter of the wrong type
    """
    def __init__(self, parameter: str):
        PowerAPIException.__init__(self)
        self.parameter = parameter


class Sync():
    """
    Sync class

    Class that receive asynchronous data and synchronize them using their timestamp
    """
    def __init__(self, type1, type2, delay):
        """
        type1 : Report -> bool.
        Function that return true if the report can be identified as the first type of report

        type2 : Report -> bool.
        Function that return true if the report can be identified as the second type of report

        delay : float
        maximal delay, in second, that is allowed between two report to pair them
        """

        if not isinstance(delay, datetime.timedelta):
            raise WrongTypeParameter("delay")

        self.delay = delay
        self.type1 = type1
        self.type2 = type2
        self.type1_buff = []
        self.type2_buff = []
        self.pair_ready = []

    def insert_report(self, report, main_buff, secondary_buff):
        """
        Insert report in the buff and delete obsolete one
        If a pair is found store it it the dedicated buff
        """
        second_report = main_buff[0]
        diff = abs(report.timestamp - second_report.timestamp)

        while diff > self.delay:
            if report.timestamp > second_report.timestamp:
                main_buff.pop(0)

            if report.timestamp < second_report.timestamp:
                return None

            if len(main_buff) == 0:
                secondary_buff.append(report)
                return None

            second_report = main_buff[0]
            diff = abs(report.timestamp - second_report.timestamp)

        if self.type1(report):
            self.pair_ready.append((report, second_report))  # report are in order (type1,type2)
        else:
            self.pair_ready.append((second_report, report))

    def add_report(self, report):
        """
        Receive a new report.
        If there is a report that can be paired with it then it do so
        Else the report is stored in the buffer
        """

        if report in self.type2_buff or report in self.type2_buff:
            logging.error("Duplicate message")
            sys.exit(0)

        if self.type1(report):
            if len(self.type2_buff) == 0:
                self.type1_buff.append(report)
            else:
                self.insert_report(report, self.type2_buff, self.type1_buff)
        elif self.type2(report):
            if len(self.type1_buff) == 0:
                self.type2_buff.append(report)
            else:
                self.insert_report(report, self.type1_buff, self.type2_buff)
        else:
            raise WrongFormatReport(type(report))

    def request(self):
        """
            Request a pair of report
            Return None if there is no pair available
        """
        if len(self.pair_ready) > 0:
            return self.pair_ready.pop(0)
        return None



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import logging
try:
    import pymongo
except ImportError:
    logging.getLogger().info("PyMongo is not installed.")

from typing import List, Type
from powerapi.database.base_db import BaseDB, DBError, IterDB
from powerapi.report import Report


class MongoBadDBError(DBError):
    """
    Error raised when hostname/port fail
    """
    def __init__(self, hostname):
        DBError.__init__(self, 'Mongo DB error : can\'t connect to ' + hostname)


class MongoIterDB(IterDB):
    """
    MongoIterDB class

    Class for iterating in a MongoDB class
    """

    def __init__(self, db, report_type, stream_mode):
        """
        """
        IterDB.__init__(self, db, report_type, stream_mode)

        #: (pymongo.Cursor): Cursor which return data
        self.cursor = None

        self.__iter__()

    def __iter__(self):
        """
        Create the iterator for get the data
        """
        if not self.stream_mode:
            self.cursor = self.db.collection.find({})
        return self

    def __next__(self) -> Report:
        """
        Allow to get the next data
        :raise: StopIteration in stream mode when no report was found. In non stream mode, raise StopIteration if the database is empty
        """
        if not self.stream_mode:
            json = self.cursor.next()
        else:
            json = self.db.collection.find_one_and_delete({})
            if json is None:
                raise StopIteration()

        return self.report_type.from_mongodb(json)


class MongoDB(BaseDB):
    """
    MongoDB class herited from BaseDB

    Allow to handle a MongoDB database in reading or writing.
    """

    def __init__(self, report_type: Type[Report], uri: str, db_name: str, collection_name: str):
        """
        :param report_type:        Type of the report handled by this database
        :param uri:             URI of the MongoDB server

        :param db_name:         database name in the mongodb
                                (ex: "powerapi")

        :param collection_name: collection name in the mongodb
                                    (ex: "sensor")
        """
        BaseDB.__init__(self, report_type)

        #: (str): URI of the mongodb server
        self.uri = uri

        #: (str): Database name in the mongodb
        self.db_name = db_name

        #: (str): Collection name in the mongodb
        self.collection_name = collection_name

        #: (pymongo.MongoClient): MongoClient instance of the server
        self.mongo_client = None

        #: (pymongo.MongoClient): MongoClient pointed to the
        #: targeted collection
        self.collection = None

        self.exceptions = [pymongo.errors.PyMongoError]

    def connect(self):
        """
        Override from BaseDB.

        It create the connection to the mongodb database with the current
        configuration (hostname/port/db_name/collection_name), then check
        if the connection has been created without failure.
        """

        # close connection if reload
        if self.mongo_client is not None:
            self.mongo_client.close()

        self.mongo_client = pymongo.MongoClient(self.uri, serverSelectionTimeoutMS=5)

        # Check if hostname:port work
        try:
            self.mongo_client.admin.command('ismaster')
        except pymongo.errors.ServerSelectionTimeoutError as exn:
            raise MongoBadDBError(self.uri) from exn

        self.collection = self.mongo_client[self.db_name][self.collection_name]

    def iter(self, stream_mode: bool) -> MongoIterDB:
        """
        Create the iterator for get the data
        """
        return MongoIterDB(self, self.report_type, stream_mode)

    def save(self, report: Report):
        """
        Override from BaseDB

        :param report: Report to save
        """
        self.collection.insert_one(self.report_type.to_mongodb(report))

    def save_many(self, reports: List[Report]):
        """
        Allow to save a batch of data

        :param reports: Batch of data.
        """
        serialized_reports = list(map(self.report_type.to_mongodb, reports))
        self.collection.insert_many(serialized_reports)



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from enum import Enum

from powerapi.report import Report


class _State(Enum):
    INIT = 1
    BLOCKED_INTER_1 = 2
    BLOCKED_INTER_2 = 3
    BLOCKED = 4
    FINAL = 5


class BlockingDetector:
    """
    Service defined for each formula that detect if the formula is blocked or not.
    A formula is defined as blocked if a three message sequence raise an exception when they are received by the formula.
    """
    def __init__(self):
        self.state = _State.INIT
        self.last_poison_message_id = None
        self.max_id_value = 10000
        self.last_message_id = 0

    def notify_poison_received(self, message: Report):
        """
        notify the blocking detector that a poison message was received
        :param message: message that make the formula crash
        """
        if self.state == _State.INIT:
            self.state = _State.BLOCKED_INTER_1
        elif message.dispatcher_report_id == self.last_poison_message_id + 1:
            if self.state == _State.BLOCKED_INTER_1:
                self.state = _State.BLOCKED_INTER_2
            elif self.state == _State.BLOCKED_INTER_2:
                self.state = _State.BLOCKED
            elif self.state == _State.BLOCKED:
                self.state = _State.FINAL
        elif self.state not in (_State.BLOCKED, _State.FINAL):
            self.state = _State.BLOCKED_INTER_1

        if message.dispatcher_report_id == self.max_id_value:
            self.last_poison_message_id = -1
        else:
            self.last_poison_message_id = message.dispatcher_report_id

    def is_blocked(self) -> bool:
        """
        :return: True if the formula is considered as blocked
        """
        return self.state == _State.BLOCKED

    def get_message_id(self) -> int:
        """
        :return: a new id for message send to formula in order to keep message order
        """
        message_id = self.last_message_id
        if self.last_message_id == self.max_id_value:
            self.last_message_id = 0
        else:
            self.last_message_id += 1
        return message_id



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import datetime

from powerapi.report import PowerReport

SENSOR_NAME = 'sensor_test'
TARGET_NAME = 'target_test'

POWER_REPORT_1 = PowerReport(0, SENSOR_NAME, TARGET_NAME, 1234, {'socket': 0, 'metadata1': 'truc', 'metadata2': 'oui'})


def gen_json_power_report(number_of_reports):
    """
    generate number_of_reports power report with json format
    each power report have the same sensor, target and power value (100W)
    only the timestamp is different.
    timestamp is generated from 01/01/1970 0:00.0 for the first report and is incremented by 1s for each returned report
    """
    reports = []
    for ts in range(number_of_reports):
        tmstp_tmp = str(datetime.datetime.fromtimestamp(ts))
        tmstp = tmstp_tmp[:tmstp_tmp.index(' ')] + 'T' + tmstp_tmp[tmstp_tmp.index(' ') + 1:] + '.0'
        reports.append(
            {
                "timestamp": tmstp,
                "sensor": SENSOR_NAME,
                "target": TARGET_NAME,
                "power": 42,
            }
        )
    return reports



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import pytest

from thespian.actors import ActorSystem, ActorExitRequest, ActorAddress

from powerapi.report import Report
from powerapi.pusher import PusherActor
from powerapi.puller import PullerActor
from powerapi.dispatcher import DispatcherActor
from powerapi.formula import FormulaActor
from powerapi.message import PusherStartMessage, PullerStartMessage, DispatcherStartMessage, FormulaStartMessage
from powerapi.message import PingMessage, OKMessage
from powerapi.supervisor import LOG_DEF


PUSHER_NAME = 'test_pusher'
PULLER_NAME = 'test_puller'
DISPATCHER_NAME = 'test_dispatcher'
DISPATCHER_DEVICE_ID = 'test_device'


@pytest.fixture
def system():
    """
    fixture that start an Actor system with log enabled before launching the test and shutdown it after the test end
    """
    syst = ActorSystem(systemBase='multiprocQueueBase', logDefs=LOG_DEF)
    yield syst
    syst.shutdown()


@pytest.fixture
def shutdown_system():
    """
    fixture that shutdown all multiproQeueuBase actor system after the test end
    """
    yield None
    syst = ActorSystem(systemBase='multiprocQueueBase', logDefs=LOG_DEF)
    syst.shutdown()


@pytest.fixture()
def puller(system):
    """
    fixture that create a PullerActor before launching the test and stop it after the test end
    """
    actor = system.createActor(PullerActor)
    yield actor
    system.tell(actor, ActorExitRequest())


@pytest.fixture()
def puller_start_message(database, report_filter, stream_mode):
    """
    return a puller start message
    """
    return PullerStartMessage('system', PULLER_NAME, database, report_filter, stream_mode)


@pytest.fixture()
def started_puller(system, puller, puller_start_message):
    """
    fixture that create and start a PullerActor and actor before launching the test and stop it after the test end
    """
    system.ask(puller, puller_start_message)
    return puller


@pytest.fixture()
def pusher(system):
    """
    fixture that create a PusherActor before launching the test and stop it after the test end
    """
    actor = system.createActor(PusherActor)
    yield actor
    system.tell(actor, ActorExitRequest())


@pytest.fixture()
def pusher_start_message(database):
    """
    return a pusher start message
    """
    return PusherStartMessage('system', PUSHER_NAME, database)


@pytest.fixture()
def started_pusher(system, pusher, pusher_start_message):
    """
    fixture that create and start a PusherActor and actor before launching the test and stop it after the test end
    """
    system.ask(pusher, pusher_start_message)
    return pusher


@pytest.fixture()
def dispatcher(system):
    """
    fixture that create a DispatcherActor before launching the test and stop it after the test end
    """
    actor = system.createActor(DispatcherActor)
    yield actor
    system.tell(actor, ActorExitRequest())


@pytest.fixture()
def dispatcher_start_message(formula_class, formula_values, route_table):
    """
    return a dispatcher start message
    """
    return DispatcherStartMessage('system', DISPATCHER_NAME, formula_class, formula_values, route_table, DISPATCHER_DEVICE_ID)


@pytest.fixture()
def started_dispatcher(system, dispatcher, dispatcher_start_message):
    """
    fixture that create and start a DispatcherActor and actor before launching the test and stop it after the test end
    """
    system.ask(dispatcher, dispatcher_start_message)
    return dispatcher


class CrashFormula(FormulaActor):
    """
    Formula that crash when receiving a report
    """
    def __init__(self):
        FormulaActor.__init__(self, FormulaStartMessage)

    @staticmethod
    def receiveMsg_Report(message: Report, sender: ActorAddress):
        """
        crash when receiving a report
        """
        raise Exception()


def is_actor_alive(system, actor_address, time=1):
    """
    wait the actor to terminate or 0.5 secondes and return its is_alive value
    """
    msg = system.ask(actor_address, PingMessage('system'), time)
    return isinstance(msg, OKMessage)



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from copy import deepcopy
from functools import reduce


class Tree:
    """
    Encapsulate the tree data structure
    """

    def __init__(self):
        self.root = None

    def add(self, path, value):
        """
        Add a leaf to the tree

        :param path: path to the node, its length must be equal to the depth
                          of the tree and the last label of the path will be the
                          label of the leaf
        :type path: list
        :param val: The value that will be stored in the leaf
        """
        if len(path) == 0:
            raise ValueError()

        if len(path) == 1:
            self.root = Node(path[0], value)
            return

        if self.root is None:
            self.root = Node(path[0])

        self.root.add_leaf(path, value)

    def get(self, path):
        """
        Retrieves all leafs value under the node designating by path

        :param path:
        :type path: list
        :rtype: list : list of leafs value
        """
        if self.root is None:
            return []
        if path == []:
            return self.root.retrieve_leaf_values([self.root.label])

        return self.root.retrieve_leaf_values(path)

    def leafs(self):
        """
        Return a list of all (path, leaf)

        :rtype: list
        """
        return self.root.get_childs()


class Node:
    """
    A labeled tree data structure that store data in its leaf

    each node is labeled with a value a node can also be a leaf so it contains
    an other value and doesn't have child node

    leafs are stored at the same depth leaf can be retrieved by using the list
    of label from a root node to the leaf for example in the folowing tree
    (where node are labeled with a letter), we can access to Leaf5 with its
    path [A,C,H]

    all the leaf of the node can be retrieved by using the path to this node,
    for example to retrieve Leaf1 and Leaf2 we use the path [A,B]. To retrieve
    all the Leaf of the tree we use the path [A]

    .. raw:: html

        <pre>
                __________Root:A _____
               |                      |
               |                      |
          __Node1:B __            _ Node2:C ____
         |            |          |      |       |
         |            |          |      |       |
        Leaf1:D   Leaf2:E   Leaf3:F   Leaf4:G   Leaf5:H
        </pre>

    Label could be any python value

    """

    def __init__(self, label, val=None):
        self.label = label
        self.is_leaf = (val is not None)

        self.childs = []
        self.val = val

    def get_childs(self):
        """
        Return all (path, value) under this node

        :rtype: list
        """
        if self.is_leaf:
            return [([self.label], self.val)]

        result = []
        for node in self.childs:
            for path, val in node.get_childs():
                result.append(([self.label] + path, val))
        return result

    def add_leaf(self, path, val):
        """
        Add a leaf to the tree

        create unexistant node between the root node and the new leaf

        :param list path: path to the node, its length must be equal to the
                          depth of the tree and the last label of the path
                          will be the label of the leaf
        :param val:       the value that will be stored in the leaf
        """
        def aux(node, depth):
            label = path[depth]
            # if node is the leaf parent, create the leaf and add it to its
            # parent
            if depth == (len(path) - 1):
                node.childs.append(Node(label, val=val))
            # otherwise find the next node in the path and go down in the tree
            else:
                child_found = False
                for child in node.childs:
                    if child.label == label:
                        aux(child, depth + 1)
                        child_found = True
                        break
                # if no intermediate node, create it
                if not child_found:
                    child = Node(label)
                    node.childs.append(child)
                    aux(child, depth + 1)

        aux(self, 1)

    def retrieve_leaf_values(self, path):
        """retrieves all leafs value under the node designating by path

        :param path:
        :type path: list
        :rtype: list : list of leafs value
        """
        def aux(node, depth):
            label = path[depth]
            # if the current node is not in the path go back to the upper node
            if label != node.label:
                return []
            # if the current node is a leaf return its value
            if depth == (len(path) - 1):
                return node._get_leafs()

            # go down in all child nodes
            return reduce(lambda acc, child: acc + aux(child, depth + 1), node.childs, [])
        return aux(self, 0)

    def _get_leafs(self):
        """
        Retrives all leafs under this node
        """
        if self.is_leaf:
            return [self.val]
        # concat all leafs value of the node's childs
        return reduce(lambda acc, child: acc + child._get_leafs(), self.childs,
                      [])

    def __eq__(self, other):
        if not isinstance(other, Node):
            return False
        if (self.label != other.label or self.val != other.val or self.is_leaf != other.is_leaf):
            return False
        sorted_child = deepcopy(self.childs)
        sorted_child.sort(key=lambda node: node.label)
        sorted_child2 = deepcopy(other.childs)
        sorted_child2.sort(key=lambda node: node.label)
        for node, other_node in zip(sorted_child, sorted_child2):
            if node != other_node:
                return False
        return True



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations


from datetime import datetime

from typing import Dict, Any

from powerapi.report.report import Report, BadInputData, CSV_HEADER_COMMON, CsvLines


CSV_HEADER_PROCFS = CSV_HEADER_COMMON + ['socket', 'cpu', 'usage', 'global_cpu_usage']


class ProcfsReport(Report):
    """
    ProcfsReport class
    JSON Procfs format
    {
    timestamp: int
    sensor: str,
    target: str,
    usage: [ {cgroup_name: str, usage: float} ]
    }


    """

    def __init__(self, timestamp: datetime, sensor: str, target: str, usage: Dict, global_cpu_usage: float, metadata: Dict[str, Any] = {}):
        """
        Initialize an Procfs report using the given parameters.
        :param datetime timestamp: Timestamp of the report
        :param str sensor: Sensor name
        :param str target: Target name
        :param Dict[str,float] usage : CGroup name and cpu_usage
        :param float global_cpu_usage : The global CPU usage, with untracked process
        """
        Report.__init__(self, timestamp, sensor, target, metadata)

        #: (dict): Events groups
        self.usage = usage
        self.global_cpu_usage = global_cpu_usage

    def __repr__(self) -> str:
        return 'ProcfsReport(%s, %s, %s, %s, %s)' % (self.timestamp, self.sensor, self.target, sorted(self.usage.keys()), str(self.metadata))

    @staticmethod
    def from_json(data: Dict) -> ProcfsReport:
        """
        Generate a report using the given data.
        :param data: Dictionary containing the report attributes
        :return: The Procfs report initialized with the given data
        """
        try:
            ts = Report._extract_timestamp(data['timestamp'])
            metadata = {} if 'metadata' not in data else data['metadata']
            return ProcfsReport(ts, data['sensor'], data['target'], data['usage'], data['global_cpu_usage'], metadata)
        except KeyError as exn:
            raise BadInputData('no field ' + str(exn.args[0]) + ' in json document', data) from exn
        except ValueError as exn:
            raise BadInputData(exn.args[0], data) from exn

    @staticmethod
    def to_json(report: ProcfsReport) -> Dict:
        return report.__dict__

    @staticmethod
    def from_mongodb(data: Dict) -> ProcfsReport:
        """ Extract a PorcfsReport fropm a mongo DB"""
        return ProcfsReport.from_json(data)

    @staticmethod
    def to_mongodb(report: ProcfsReport) -> Dict:
        """ Export a ProcfsReport to a mongo DB"""
        return ProcfsReport.to_json(report)

    @staticmethod
    def from_csv_lines(lines: CsvLines) -> ProcfsReport:
        """ Extract a ProcfsReport from a csv"""
        sensor_name = None
        target = None
        timestamp = None
        usage = {}
        global_cpu_usage = None
        metadata = {}
        for file_name, row in lines:
            cgroup_name = file_name[:-4] if file_name[len(file_name) - 4:] == '.csv' else file_name
            try:
                if sensor_name is None:
                    sensor_name = row['sensor']
                    target = row['target']
                    timestamp = Report._extract_timestamp(row['timestamp'])
                    global_cpu_usage = row['global_cpu_usage']
                else:
                    if sensor_name != row['sensor']:
                        raise BadInputData('csv line with different sensor name are mixed into one report', {})
                    if target != row['target']:
                        raise BadInputData('csv line with different target are mixed into one report', {})
                    if timestamp != ProcfsReport._extract_timestamp(row['timestamp']):
                        raise BadInputData('csv line with different timestamp are mixed into one report', {})
                    if global_cpu_usage != row['global_cpu_usage']:
                        raise BadInputData('Different cpu usage are provided in one report', {})

                ProcfsReport._create_cgroup(row, cgroup_name, usage)

                for key, value in row.items():
                    if key not in CSV_HEADER_PROCFS:
                        metadata[key] = value

            except KeyError as exn:
                raise BadInputData('missing field ' + str(exn.args[0]) + ' in csv file ' + file_name, {}) from exn
            except ValueError as exn:
                raise BadInputData(exn.args[0], row) from exn

        return ProcfsReport(timestamp, sensor_name, target, usage, global_cpu_usage, metadata)

    @staticmethod
    def _create_cgroup(_, cgroup_name, usage):
        if cgroup_name not in usage:
            usage[cgroup_name] = {}


def create_report_root(cgroup_list, timestamp=datetime.fromtimestamp(0), sensor='toto', target='all'):
    """ Create a default procfs report """
    sensor = ProcfsReport(timestamp=timestamp, sensor=sensor, target=target, usage={}, global_cpu_usage=0)
    for (cgroup_id, cpu_usage) in cgroup_list:
        sensor.usage[cgroup_id] = cpu_usage
    return sensor



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

from datetime import datetime
from typing import Dict, Any, List, Tuple

from powerapi.report.report import Report, CSV_HEADER_COMMON, BadInputData, CsvLines


CSV_HEADER_POWER = CSV_HEADER_COMMON + ['power', 'socket']


class PowerReport(Report):
    """
    PowerReport stores the power estimation information.
    """

    def __init__(self, timestamp: datetime, sensor: str, target: str, power: float, metadata: Dict[str, Any] = {}):
        """
        Initialize a Power report using the given parameters.
        :param datetime timestamp: Report timestamp
        :param str sensor: Sensor name
        :param str target: Target name
        :param float power: Power value
        :param dict metadata: Metadata values, can be anything that add useful information
        """
        Report.__init__(self, timestamp, sensor, target, metadata)

        self.power = power

    def __repr__(self) -> str:
        return 'PowerReport(%s, %s, %s, %f, %s)' % (self.timestamp, self.sensor, self.target, self.power, str(self.metadata))

    @staticmethod
    def from_json(data: Dict) -> Report:
        """
        Generate a report using the given data.
        :param data: Dictionary containing the report attributes
        :return: The HWPC report initialized with the given data
        """
        try:
            ts = Report._extract_timestamp(data['timestamp'])
            metadata = {} if 'metadata' not in data else data['metadata']
            return PowerReport(ts, data['sensor'], data['target'], data['power'], metadata)
        except KeyError as exn:
            raise BadInputData('PowerReport require field ' + str(exn.args[0]) + ' in json document', data) from exn
        except ValueError as exn:
            raise BadInputData(exn.args[0], data) from exn

    @staticmethod
    def from_csv_lines(lines: CsvLines) -> PowerReport:
        """
        :param lines: list of pre-parsed lines. a line is a tuple composed with :
                         - the file name where the line were read
                         - a dictionary where key is column name and value is the value read from the line
        :return: a PowerReport that contains value from the given lines
        """
        if len(lines) != 1:
            raise BadInputData('a power report could only be parsed from one csv line', None)
        file_name, row = lines[0]

        try:
            sensor_name = row['sensor']
            target = row['target']
            timestamp = Report._extract_timestamp(row['timestamp'])
            power = float(row['power'])
            metadata = {}

            for key in row.keys():
                if key not in CSV_HEADER_POWER:
                    metadata[key] = row[key]
            return PowerReport(timestamp, sensor_name, target, power, metadata)

        except KeyError as exn:
            raise BadInputData('missing field ' + str(exn.args[0]) + ' in csv file ' + file_name, row) from exn
        except ValueError as exn:
            raise BadInputData(exn.args[0], row) from exn

    @staticmethod
    def to_csv_lines(report: PowerReport, tags: List[str]) -> CsvLines:
        """
        convert a power report into csv lines
        :param report: Report that will be converted into csv lines
        :param tags: metadata added as columns in csv file
        :return: list of pre-parsed lines. a line is a tuple composed with :
                   - the file name where the line were read
                   - a dictionary where key is column name and value is the value read from the line
        """
        line = {
            'sensor': report.sensor,
            'target': report.target,
            'timestamp': int(datetime.timestamp(report.timestamp) * 1000),
            'power': report.power
        }

        # Copy all metadata
        for tag, value in report.metadata.items():
            line[tag] = value

        # Check that attended metadata are there
        for tag in tags:
            if tag not in report.metadata:
                raise BadInputData('no tag ' + tag + ' in power report', report)
            line[tag] = report.metadata[tag]

        final_dict = {'PowerReport': [line]}
        return CSV_HEADER_POWER, final_dict

    @staticmethod
    def to_virtiofs_db(report: PowerReport) -> Tuple[str, str]:
        """
        return a tuple containing the power value and the name of the file to store the value.
        """
        if 'socket' not in report:
            raise BadInputData('no tag socket in power report', report)
        filename = 'power_consumption_package' + str(report['socket'])
        power = report.power
        return filename, power

    def _gen_tag(self, metadata_keept):
        tags = {'sensor': self.sensor,
                'target': self.target
                }

        for metadata_name in metadata_keept:
            if metadata_name not in self.metadata:
                raise BadInputData('no tag ' + metadata_name + ' in power report', self)
            else:
                tags[metadata_name] = self.metadata[metadata_name]

        return tags

    @staticmethod
    def to_influxdb(report: PowerReport, tags: List[str]) -> Dict:
        """
        :return: a dictionary, that can be stored into an influxdb, from a given PowerReport
        """
        return {
            'measurement': 'power_consumption',
            'tags': report._gen_tag(tags),
            'time': str(report.timestamp),
            'fields': {
                'power': report.power
            }
        }

    @staticmethod
    def to_prometheus(report: PowerReport, tags: List[str]) -> Dict:
        """
        :return: a dictionary, that can be stored into a prometheus instance, from a given PowerReport
        """
        return {
            'tags': report._gen_tag(tags),
            'time': int(report.timestamp.timestamp()),
            'value': report.power
        }

    @staticmethod
    def to_mongodb(report: PowerReport) -> Dict:
        """
        :return: a dictionary, that can be stored into a mongodb, from a given PowerReport
        """
        return PowerReport.to_json(report)

    @staticmethod
    def from_mongodb(data: Dict) -> Report:
        """
        :return: a PowerReport from a dictionary pulled from mongodb
        """
        return PowerReport.from_json(data)



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import sys

from powerapi.cli.config_parser import MainConfigParser, SubConfigParser
from powerapi.cli.parser import store_true
from powerapi.cli.parser import MissingValueException
from powerapi.cli.parser import BadTypeException, BadContextException
from powerapi.cli.parser import UnknowArgException


def extract_file_names(arg, val, args, acc):
    """
    action used to convert string from --files parameter into a list of file name
    """
    acc[arg] = val.split(",")
    return args, acc


class CommonCLIParser(MainConfigParser):
    """
    PowerAPI basic config parser
    """

    def __init__(self):
        MainConfigParser.__init__(self)

        self.add_argument(
            "v",
            "verbose",
            flag=True,
            action=store_true,
            default=False,
            help="enable verbose mode",
        )
        self.add_argument(
            "s",
            "stream",
            flag=True,
            action=store_true,
            default=False,
            help="enable stream mode",
        )

        subparser_libvirt_mapper_modifier = SubConfigParser("libvirt_mapper")
        subparser_libvirt_mapper_modifier.add_argument(
            "u", "uri", help="libvirt daemon uri", default=""
        )
        subparser_libvirt_mapper_modifier.add_argument(
            "d",
            "domain_regexp",
            help="regexp used to extract domain from cgroup string",
        )
        subparser_libvirt_mapper_modifier.add_argument("n", "name", help="")
        self.add_subparser(
            "report_modifier",
            subparser_libvirt_mapper_modifier,
            help="Specify a report modifier to change input report values : --report_modifier ARG1 ARG2 ...",
        )

        subparser_mongo_input = SubConfigParser("mongodb")
        subparser_mongo_input.add_argument("u", "uri", help="specify MongoDB uri")
        subparser_mongo_input.add_argument(
            "d",
            "db",
            help="specify MongoDB database name",
        )
        subparser_mongo_input.add_argument(
            "c", "collection", help="specify MongoDB database collection"
        )
        subparser_mongo_input.add_argument(
            "n", "name", help="specify puller name", default="puller_mongodb"
        )
        subparser_mongo_input.add_argument(
            "m",
            "model",
            help="specify data type that will be storen in the database",
            default="HWPCReport",
        )
        self.add_subparser(
            "input",
            subparser_mongo_input,
            help="specify a database input : --db_input database_name ARG1 ARG2 ... ",
        )

        subparser_socket_input = SubConfigParser("socket")
        subparser_socket_input.add_argument(
            "p", "port", type=int, help="specify port to bind the socket"
        )
        subparser_socket_input.add_argument(
            "n", "name", help="specify puller name", default="puller_socket"
        )
        subparser_socket_input.add_argument(
            "m",
            "model",
            help="specify data type that will be sent through the socket",
            default="HWPCReport",
        )
        self.add_subparser(
            "input",
            subparser_socket_input,
            help="specify a database input : --db_input database_name ARG1 ARG2 ... ",
        )

        subparser_csv_input = SubConfigParser("csv")
        subparser_csv_input.add_argument(
            "f",
            "files",
            help="specify input csv files with this format : file1,file2,file3",
            action=extract_file_names,
            default=[],
        )
        subparser_csv_input.add_argument(
            "m",
            "model",
            help="specify data type that will be storen in the database",
            default="HWPCReport",
        )
        subparser_csv_input.add_argument(
            "n", "name", help="specify puller name", default="puller_csv"
        )
        self.add_subparser(
            "input",
            subparser_csv_input,
            help="specify a database input : --db_input database_name ARG1 ARG2 ... ",
        )

        subparser_file_input = SubConfigParser("filedb")
        subparser_file_input.add_argument(
            "m",
            "model",
            help="specify data type that will be storen in the database",
            default="PowerReport",
        )
        subparser_file_input.add_argument("f", "filename", help="specify file name")
        subparser_file_input.add_argument(
            "n", "name", help="specify pusher name", default="pusher_filedb"
        )
        self.add_subparser(
            "input",
            subparser_file_input,
            help="specify a database input : --db_input database_name ARG1 ARG2 ... ",
        )

        subparser_file_output = SubConfigParser("filedb")
        subparser_file_output.add_argument(
            "m",
            "model",
            help="specify data type that will be storen in the database",
            default="PowerReport",
        )
        subparser_file_output.add_argument("f", "filename", help="specify file name")
        subparser_file_output.add_argument(
            "n", "name", help="specify pusher name", default="pusher_filedb"
        )
        self.add_subparser(
            "output",
            subparser_file_output,
            help="specify a database output : --db_output database_name ARG1 ARG2 ...",
        )

        subparser_virtiofs_output = SubConfigParser("virtiofs")
        help = "regexp used to extract vm name from report."
        help += "The regexp must match the name of the target in the HWPC-report and a group must"
        subparser_virtiofs_output.add_argument("r", "vm_name_regexp", help=help)
        subparser_virtiofs_output.add_argument(
            "d",
            "root_directory_name",
            help="directory where VM directory will be stored",
        )
        subparser_virtiofs_output.add_argument(
            "p",
            "vm_directory_name_prefix",
            help="first part of the VM directory name",
            default="",
        )
        subparser_virtiofs_output.add_argument(
            "s",
            "vm_directory_name_suffix",
            help="last part of the VM directory name",
            default="",
        )
        subparser_virtiofs_output.add_argument(
            "m",
            "model",
            help="specify data type that will be storen in the database",
            default="PowerReport",
        )
        subparser_virtiofs_output.add_argument(
            "n", "name", help="specify pusher name", default="pusher_virtiofs"
        )
        self.add_subparser(
            "output",
            subparser_virtiofs_output,
            help="specify a database output : --db_output database_name ARG1 ARG2 ...",
        )

        subparser_mongo_output = SubConfigParser("mongodb")
        subparser_mongo_output.add_argument("u", "uri", help="specify MongoDB uri")
        subparser_mongo_output.add_argument(
            "d", "db", help="specify MongoDB database name"
        )
        subparser_mongo_output.add_argument(
            "c", "collection", help="specify MongoDB database collection"
        )

        subparser_mongo_output.add_argument(
            "m",
            "model",
            help="specify data type that will be storen in the database",
            default="PowerReport",
        )
        subparser_mongo_output.add_argument(
            "n", "name", help="specify pusher name", default="pusher_mongodb"
        )
        self.add_subparser(
            "output",
            subparser_mongo_output,
            help="specify a database output : --db_output database_name ARG1 ARG2 ...",
        )

        subparser_prom_output = SubConfigParser("prom")
        subparser_prom_output.add_argument("t", "tags", help="specify report tags")
        subparser_prom_output.add_argument("u", "uri", help="specify server uri")
        subparser_prom_output.add_argument(
            "p", "port", help="specify server port", type=int
        )
        subparser_prom_output.add_argument(
            "M", "metric_name", help="speify metric name"
        )
        subparser_prom_output.add_argument(
            "d",
            "metric_description",
            help="specify metric description",
            default="energy consumption",
        )
        help = "specify number of second for the value must be aggregated before compute statistics on them"
        subparser_prom_output.add_argument(
            "A", "aggregation_period", help=help, default=15, type=int
        )

        subparser_prom_output.add_argument(
            "m",
            "model",
            help="specify data type that will be storen in the database",
            default="PowerReport",
        )
        subparser_prom_output.add_argument(
            "n", "name", help="specify pusher name", default="pusher_prom"
        )
        self.add_subparser(
            "output",
            subparser_prom_output,
            help="specify a database output : --db_output database_name ARG1 ARG2 ...",
        )

        subparser_direct_prom_output = SubConfigParser("direct_prom")
        subparser_direct_prom_output.add_argument(
            "t", "tags", help="specify report tags"
        )
        subparser_direct_prom_output.add_argument("a", "uri", help="specify server uri")
        subparser_direct_prom_output.add_argument(
            "p", "port", help="specify server port", type=int
        )
        subparser_direct_prom_output.add_argument(
            "M", "metric_name", help="speify metric name"
        )
        subparser_direct_prom_output.add_argument(
            "d",
            "metric_description",
            help="specify metric description",
            default="energy consumption",
        )
        subparser_direct_prom_output.add_argument(
            "m",
            "model",
            help="specify data type that will be storen in the database",
            default="PowerReport",
        )
        subparser_direct_prom_output.add_argument(
            "n", "name", help="specify pusher name", default="pusher_prom"
        )
        self.add_subparser(
            "output",
            subparser_direct_prom_output,
            help="specify a database output : --db_output database_name ARG1 ARG2 ...",
        )

        subparser_csv_output = SubConfigParser("csv")
        subparser_csv_output.add_argument(
            "d",
            "directory",
            help="specify directory where where output  csv files will be writen",
        )
        subparser_csv_output.add_argument(
            "m",
            "model",
            help="specify data type that will be storen in the database",
            default="PowerReport",
        )

        subparser_csv_output.add_argument("t", "tags", help="specify report tags")
        subparser_csv_output.add_argument(
            "n", "name", help="specify pusher name", default="pusher_csv"
        )
        self.add_subparser(
            "output",
            subparser_csv_output,
            help="specify a database outpout : --db_output database_name ARG1 ARG2 ... ",
        )

        subparser_influx_output = SubConfigParser("influxdb")
        subparser_influx_output.add_argument("u", "uri", help="specify InfluxDB uri")
        subparser_influx_output.add_argument("t", "tags", help="specify report tags")
        subparser_influx_output.add_argument(
            "d", "db", help="specify InfluxDB database name"
        )
        subparser_influx_output.add_argument(
            "p", "port", help="specify InfluxDB connection port", type=int
        )
        subparser_influx_output.add_argument(
            "m",
            "model",
            help="specify data type that will be storen in the database",
            default="PowerReport",
        )
        subparser_influx_output.add_argument(
            "n", "name", help="specify pusher name", default="pusher_influxdb"
        )
        self.add_subparser(
            "output",
            subparser_influx_output,
            help="specify a database output : --db_output database_name ARG1 ARG2 ... ",
        )

        subparser_opentsdb_output = SubConfigParser("opentsdb")
        subparser_opentsdb_output.add_argument("u", "uri", help="specify openTSDB host")
        subparser_opentsdb_output.add_argument(
            "p", "port", help="specify openTSDB connection port", type=int
        )
        subparser_opentsdb_output.add_argument(
            "metric_name", help="specify metric name"
        )

        subparser_opentsdb_output.add_argument(
            "m",
            "model",
            help="specify data type that will be storen in the database",
            default="PowerReport",
        )
        subparser_opentsdb_output.add_argument(
            "n", "name", help="specify pusher name", default="pusher_opentsdb"
        )
        self.add_subparser(
            "output",
            subparser_opentsdb_output,
            help="specify a database output : --db_output database_name ARG1 ARG2 ... ",
        )

    def parse_argv(self):
        """ """
        try:
            return self.parse(sys.argv[1:])

        except MissingValueException as exn:
            msg = "CLI error : argument " + exn.argument_name + " : expect a value"
            print(msg, file=sys.stderr)

        except BadTypeException as exn:
            msg = "Configuration error : " + exn.msg
            print(msg, file=sys.stderr)

        except UnknowArgException as exn:
            msg = "CLI error : unknow argument " + exn.argument_name
            print(msg, file=sys.stderr)

        except BadContextException as exn:
            msg = "CLI error : argument " + exn.argument_name
            msg += " not used in the correct context\nUse it with the following arguments :"
            for main_arg_name, context_name in exn.context_list:
                msg += "\n  --" + main_arg_name + " " + context_name
            print(msg, file=sys.stderr)

        sys.exit()



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import logging
from typing import List, Type
try:
    from influxdb import InfluxDBClient
    from requests.exceptions import ConnectionError as InfluxConnectionError
except ImportError:
    logging.getLogger().info("influx-client is not installed.")

from powerapi.report import Report
from .base_db import BaseDB, DBError


class CantConnectToInfluxDBException(DBError):
    """
    Exception raised to notify that connection to the influx database is impossible
    """


class InfluxDB(BaseDB):
    """
    MongoDB class herited from BaseDB

    Allow to handle a InfluxDB database in reading or writing.
    """

    def __init__(self, report_type: Type[Report], uri: str, port, db_name: str, tags: List[str]):
        """
        :param url:             URL of the InfluxDB server
        :param port:            port of the InfluxDB server

        :param db_name:         database name in the influxdb
                                    (ex: "powerapi")

        :param report_type:        Type of the report handled by this database
        :param tags: metadata used to tag metric

        """
        BaseDB.__init__(self, report_type)
        self.uri = uri
        self.port = port
        self.db_name = db_name
        self.tags = tags

        self.client = None

    def __iter__(self):
        raise NotImplementedError()

    def _ping_client(self):
        if hasattr(self.client, 'ping'):
            self.client.ping()
        else:
            self.client.request(url="ping", method='GET', expected_response_code=204)

    def connect(self):
        """
        Override from BaseDB.

        Create the connection to the influxdb database with the current
        configuration (hostname/port/db_name), then check if the connection has
        been created without failure.

        """
        # close connection if reload
        if self.client is not None:
            self.client.close()

        self.client = InfluxDBClient(host=self.uri, port=self.port, database=self.db_name)
        try:
            self._ping_client()
        except InfluxConnectionError as exn:
            raise CantConnectToInfluxDBException('connexion error') from exn

        for db in self.client.get_list_database():
            if db['name'] == self.db_name:
                return

        self.client.create_database(self.db_name)

    def save(self, report: Report):
        """
        Override from BaseDB

        :param report: Report to save
        """
        data = self.report_type.to_influxdb(report, self.tags)
        for tag in data['tags']:
            data['tags'][tag] = str(data['tags'][tag])
        self.client.write_points([data])

    def save_many(self, reports: List[Report]):
        """
        Save a batch of data

        :param reports: Batch of data.
        """

        data_list = list(map(lambda r: self.report_type.to_influxdb(r, self.tags), reports))
        self.client.write_points(data_list)



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
try:
    from libvirt import libvirtError
except ImportError:
    from powerapi.report_modifier.libvirt_mapper import LibvirtException
    libvirtError = LibvirtException

DOMAIN_NAME_1 = 'instance-00000001'
DOMAIN_NAME_2 = 'instance-00000999'

LIBVIRT_TARGET_NAME1 = '/machine/qemu-1-' + DOMAIN_NAME_1 + '.libvirt-qemu'
LIBVIRT_TARGET_NAME2 = '/machine/qemu-2-' + DOMAIN_NAME_2 + '.libvirt-qemu'

REGEXP = '/machine/qemu-[0-9]+-(.*)\\.libvirt-qemu'

UUID_1 = '123-456-789-1011'


class MockedDomain:
    """
    Mocked libvirt domain that contain only uuid string
    """
    def __init__(self, uuid_str):
        self.uuid_str = uuid_str

    def UUIDString(self):
        """
        mocked UUIDString method
        """
        return self.uuid_str


class MockedLibvirt:
    """
    Mocked libvirt client
    """
    def lookupByName(self, domain_name):
        """
        :return: MockedDomain with UUID_1
        """
        if domain_name == DOMAIN_NAME_1:
            return MockedDomain(UUID_1)
        raise libvirtError('')



# Copyright (c) 2021, INRIA
# Copyright (c) 2021, University of Lille
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import csv
import os

from typing import List, Type
from powerapi.report.report import Report, CSV_HEADER_COMMON
from powerapi.database.base_db import BaseDB, IterDB
from powerapi.exception import PowerAPIException
from powerapi.utils import utils

# Array of field that will not be considered as a group
COMMON_ROW = ['timestamp', 'sensor', 'target', 'socket', 'cpu']


class CsvBadFilePathError(PowerAPIException):
    """
    Error raised when file is not found
    """


class CsvBadCommonKeysError(PowerAPIException):
    """
    Error raised when a common keys is not found
    """


class HeaderAreNotTheSameError(PowerAPIException):
    """
    Error raised when the header read in a file doesn't fit the input data
    """


class CsvIterDB(IterDB):
    """
    IterDB class

    This class allows to browse a database as an iterable
    """

    def __init__(self, db, filenames, report_type, stream_mode):
        """
        """
        super().__init__(db, report_type, stream_mode)

        self.filenames = filenames

        # intern memory for reading
        # path_file: {
        #     'next_line': [],
        #     'reader': None,
        #     'file': None
        # }
        self.tmp_read = {}

        # Add it in the tmp
        for filename in filenames:
            self.tmp_read[filename] = {
                'next_line': [],
                'reader': None,
                'file': None
            }

        # Open all files with csv and read first line
        for filename in self.filenames:
            try:
                self.tmp_read[filename]['file'] = open(filename)
                self.tmp_read[filename]['reader'] = csv.DictReader(self.tmp_read[filename]['file'])
            except FileNotFoundError as error:
                raise CsvBadFilePathError(error) from error
            self.tmp_read[filename]['next_line'] = self._next(filename)

            # Check common key
            for key in CSV_HEADER_COMMON:
                if key not in self.tmp_read[filename]['next_line']:
                    raise CsvBadCommonKeysError("Wrong columns keys")

        # Save the first timestamp
        if self.filenames:
            self.saved_timestamp = utils.timestamp_to_datetime(
                int(self.tmp_read[self.filenames[0]]['next_line']['timestamp']))

    def __iter__(self):
        """
        """
        return self

    def _next(self, filename):
        """
        Get next row, None otherwise

        :param str filename: file name we want to read
        """
        try:
            return self.tmp_read[filename]['reader'].__next__()
        except StopIteration:
            return None

    def _close_file(self):
        for filename in self.filenames:
            if self.tmp_read[filename]['file'] is not None:
                self.tmp_read[filename]['file'].close()

    def __next__(self) -> Report:
        """
        Allow to get the next data
        """
        raw_data = []
        # Get the current timestamp
        current_timestamp = self.saved_timestamp
        previous_target = None
        # For all files

        for path_file in self.filenames:
            # While timestamp is lower or equal
            while True:
                # Get the next line
                row = self.tmp_read[path_file]['next_line']

                # If nothing more, break
                if row is None:
                    # If the first file a no next file, just stop the iteration
                    if not raw_data and path_file == self.filenames[0]:
                        self._close_file()
                        raise StopIteration()
                    else:
                        break

                # Get the timestamp as datetime
                row_timestamp = utils.timestamp_to_datetime(
                    int(row['timestamp']))
                # If timestamp is higher, we stop here
                if row_timestamp > current_timestamp:
                    if path_file == self.filenames[-1]:
                        self.saved_timestamp = row_timestamp
                    break  # move to next file

                if row_timestamp < current_timestamp:
                    self.tmp_read[path_file]['next_line'] = self._next(path_file)
                    continue

                if previous_target is not None:
                    if row['target'] != previous_target:
                        break  # move to next file
                else:
                    previous_target = row['target']

                # Else if it's the same, we merge
                raw_data.append((path_file.split('/')[-1], row))
                # Next line
                self.tmp_read[path_file]['next_line'] = self._next(path_file)

        if not raw_data:
            self._close_file()
            raise StopIteration()

        report = self.report_type.from_csv_lines(raw_data)

        return report


class CsvDB(BaseDB):
    """
    CsvDB class herited from BaseDB

    This class define the behaviour for reading some csv file.
    a CsvDB instance can be define by its current path
    """

    def __init__(self, report_type: Type[Report], tags: List[str], current_path="/tmp/csvdbtest", files=[]):
        """
        :param current_path: Current path where read/write files
        """
        BaseDB.__init__(self, report_type)

        #: (list): list of file name .csv
        self.filenames = []

        #: (str): current path
        self.current_path = current_path if current_path[-1] == '/' else current_path + '/'

        #: (int): allow to know if we read a new report, or the same
        #: current timestamp
        self.saved_timestamp = utils.timestamp_to_datetime(0)
        self.tags = tags

        self.add_files(files)

    ##################
    # Specific CsvDB #
    ##################

    def add_file(self, filename):
        """
        Add a file in the filenames list (it can be relative or absolute path)
        :param filename: Path to file
        """
        # If absolute path
        if filename[0] == '/':
            self.filenames.append(filename)
        else:
            filename = self.current_path + filename
            self.filenames.append(filename)

    def add_files(self, filenames):
        """
        Add list of files in the filenames list (it can be relative or absolute path)
        :param filenames: List of path to file
        """
        for filename in filenames:
            self.add_file(filename)

    def clean_files(self):
        """
        Clean the filenames list
        """
        self.filenames.clear()

    ########################
    # Override from BaseDB #
    ########################

    def iter(self, stream_mode: bool) -> CsvIterDB:
        """
        Create the iterator for get the data
        """
        return CsvIterDB(self, self.filenames, self.report_type, stream_mode)

    def connect(self):
        """
        Override from BaseDB.

        Nothing to do with CSV, because it's just files operations.
        """

    def save(self, report: Report):
        """
        Allow to save a serialized_report in the db

        :param report: Report
        """
        csv_header, data = self.report_type.to_csv_lines(report, self.tags)

        # If the repository doesn't exist, create it
        rep_path = self.current_path + report.sensor + "-" + report.target
        try:
            os.makedirs(rep_path)
        except FileExistsError:
            pass

        for filename, values in data.items():
            rep_path_with_file = rep_path + '/' + filename + '.csv'

            # Get the header and check if it's ok
            for value in values:
                header = csv_header + sorted(list(set([event_key for event_key, _ in value.items()]) - set(csv_header)))
                header_exist = False
                try:
                    with open(rep_path_with_file) as csvfile:
                        reader = csv.DictReader(csvfile)
                        if reader.fieldnames:
                            header_exist = True

                        if header != reader.fieldnames:
                            raise HeaderAreNotTheSameError("Header are not the same in " + rep_path_with_file)
                        csvfile.close()
                except FileNotFoundError:
                    pass

                # Write
                with open(rep_path_with_file, 'a') as csvfile:
                    writer = csv.DictWriter(csvfile, header)
                    if not header_exist:
                        writer.writeheader()
                    writer.writerow(value)
                    csvfile.close()

    def save_many(self, reports: List[Report]):
        """
        Allow to save a batch of report

        :param reports: Batch of report.
        """
        for report in reports:
            self.save(report)
